(* $Id: gtf.ml,v 1.28 2006/12/08 10:00:34 heinz Exp $
 *
 * GTF (Generic Test Framework)
 * Copyright (C) 2004  Heinz Deinhart
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *)

let print_replace_regexps =
  [
    Str.regexp( "\\^" ); Str.regexp( "%" ); Str.regexp( "[|]" );
    Str.regexp( "\\~" ); Str.regexp( "\"" ); Str.regexp( "[&]" )
  ]
let print_replace_targets =
  [
    "$^\\wedge$"; "\\%"; "$|$";
    "$\\sim$"; "''"; "\\&";
  ]

let debug_message_level = 5

let dprintf (lvl: int) str =
  if lvl <= debug_message_level then begin
    flush_all ();
    Printf.fprintf stderr "DEBUG: %s\n" str;
    flush stderr
  end

class virtual printable = object(self)
  method virtual print : bool -> string
  method print_to b channel =
    output_string channel (self#print b)
end

class virtual archivable = object
  method virtual archive : unit -> string
  method virtual unarchive : string -> unit
end

type question_status =
    QS_Default
  | QS_Modified
  | QS_Done

class virtual question csc = object(self)
  inherit archivable
  inherit printable

  val check_status_callback: unit -> unit = csc

  method virtual classname : unit -> string
  method virtual populate_box : (GObj.widget -> unit) -> unit
  method virtual status : unit -> question_status
  method virtual max_pts: unit -> int
  method virtual act_pts: unit -> int
  method virtual min_pts: unit -> int
  method as_question () =
    (self :> question)
end

class student = object(_)
  inherit archivable

  val mutable idnr: string = ""
  val mutable lastname: string = ""
  val mutable firstname: string = ""

  method lastname () =
    lastname
  method set_lastname str =
    lastname <- str

  method firstname () =
    firstname
  method set_firstname str =
    firstname <- str

  method idnr () =
    idnr
  method set_idnr str =
    idnr <- str

  method compare_idnr str =
    if (String.compare (String.lowercase_ascii str) 
          (String.lowercase_ascii idnr)) = 0 then
      true
    else
      false

  method compare_lastname str  =
    if (String.compare (String.lowercase_ascii str)
	  (String.lowercase_ascii lastname)) = 0 then
      true
    else
      false

  method archive () =
    Marshal.to_string (idnr, lastname, firstname) []

  method unarchive str =
    match Marshal.from_string str 0 with
	( mnr, nn, vn ) ->
	  idnr <- mnr;
	  lastname <- nn;
	  firstname <- vn
end

class test (question_generator: (unit -> unit) -> string -> question) =
object(self)
  inherit archivable
  inherit printable

  val mutable questions: question list  = []
  val mutable student: student = new student
  val mutable qgen = question_generator
  val mutable shuffle = false
  val mutable time: float = 1800.0
  val mutable display_status: int -> int * int * int -> unit = fun _ _ -> ()
  val mutable last_modified: float = 0.0

  method check_status () =
    let tuple3_add (sdef,smod,sdon) = function
	QS_Default -> (sdef+1, smod, sdon)
      | QS_Modified -> (sdef, smod+1, sdon)
      | QS_Done -> (sdef, smod, sdon+1)
    in let st = List.fold_right (fun q s -> tuple3_add s (q#status ()))
		  questions (0,0,0)
    in
      display_status (List.length questions) st

  method set_display_status callback =
    display_status <- callback

  method questions () =
      questions

  method student =
    student

  method set_shuffle b =
    shuffle <- b

  method touch () =
    last_modified <- Unix.time ()

  method shuffle_questions () =
    questions <- let ar = Array.of_list questions in
      Util.array_randomize ar;
      Array.to_list ar

  method time =
    time

  method set_time secs =
    time <- secs

  method archive () =
    Marshal.to_string ((List.map (fun obj -> obj#classname (), obj#archive ())
			 questions), (student#archive ()), time, last_modified) []

  method unarchive x =
    let qarch, studarch, _time, last_mod = Marshal.from_string x 0
    in
      last_modified <- last_mod;
      time <- _time;
      student#unarchive studarch;
      questions <- List.map (fun (classname, str) ->
			       let obj = qgen self#check_status classname in
				 obj#unarchive str; obj) qarch

  method print explain =
    let print_pnts q =
      if explain then
	Printf.sprintf "\n\n\\textit{Punkte: %i von %i moeglichen}\n\n"
	  (q#act_pts ()) (q#max_pts ())
      else
	""
    in let touched = Unix.localtime last_modified
    in let rawstr =
      (Printf.sprintf "\\begin{center}{\\Huge Test fuer %s(%s) vom %i.%i.%i um %02i:%02i}\\end{center}\n" (self#student#lastname ())
	 (self#student#idnr ()) touched.Unix.tm_mday touched.Unix.tm_mon touched.Unix.tm_year
	 touched.Unix.tm_hour touched.Unix.tm_min) ^
	(List.fold_right (fun q s -> s ^ (q#print explain) ^ (print_pnts q))
	   (List.rev questions) "")
    in
      List.fold_right2 (fun r t s -> Str.global_replace r t s)
	print_replace_regexps print_replace_targets rawstr

  method print_calculation () =
    let actpts =
      List.fold_right (+) (List.map (fun q -> q#act_pts ()) questions) 0
    and maxpts =
      List.fold_right (+) (List.map (fun q -> q#max_pts ()) questions) 0
    in
      Printf.sprintf "\n\n\\vspace*{1cm}\\textit{\\huge Ergebnis:}\n\nSie haben von %i moeglichen %i Punkt(e) bekommen!\n\n" maxpts actpts;

  method populate_vbox (vbox: GPack.box) =
    let rec work i = function
	q :: rest ->
	  let frame = GBin.frame ~label:("Frage " ^ (string_of_int i))
			~packing:vbox#pack ()
	  in
	    q#populate_box frame#add;
	    work (i+1) rest;
	    ()
      | [] ->
	  ()
    in
      work 1 questions

  method add_question q =
    questions <- List.append questions [q]

  method add_questions q =
    questions <- List.append questions q
end

exception GTF_protocol_error

type gtfprotocol =
    RequestDirectory
  | Directory of string list
(* unused *)  | RequestEntries of string 
(* unused *)  | Entries of (string * string list)
  | RequestTest of (string * string * string)
  | Test of (string option * string * string * string)
  | Intermediate of (string * string * (string * string))
  | Finish of (string * string * (string * string))
  | Time of string
  | ServerError of string
  | ClientError of string

class virtual network_buffer = object (self)
  method virtual out_fd : unit -> Unix.file_descr
  method virtual in_fd : unit -> Unix.file_descr

  method send_msg (data: gtfprotocol) =
    dprintf 10 "network_buffer: send_msg called..";
    let rec send_str fd str start count =
      dprintf 10 (Printf.sprintf "network_buffer calling send_str _fd _str start=%i count=%i" start count);
      match count with
	  0 -> ()
	| _ when count < 0 ->
	    raise (Invalid_argument "send_msg: count < 0 error")
	| _ ->
	    let sent = Unix.send fd str start count [] in
	      send_str fd str (start + sent) (count - sent) in
    let str_msg = Marshal.to_string data [] in
    let str_msg_len = String.length str_msg in
      send_str (self#out_fd ()) (Bytes.of_string 
                                   (Printf.sprintf "%.10i" str_msg_len)) 0 10;
      dprintf 10 "network_buffer:    ..header sent..";
      send_str (self#out_fd ()) (Bytes.of_string str_msg) 0 str_msg_len;
      dprintf 10 (Printf.sprintf "network_buffer:    ..body (%s) sent: "
		 (Digest.to_hex (Digest.string str_msg)))

  method recv_msg () : gtfprotocol =
    dprintf 10 "network_buffer: recv_msg called..";
    let rec recv_str fd str start count =
      dprintf 10 (Printf.sprintf "network_buffer calling recv_str _fd _str start=%i count=%i" start count);
      match count with
	  0 -> ()
	| _ when count < 0 ->
	    raise (Invalid_argument "recv_msg: count < 0 error")
	| _ ->
	    let recvd = Unix.recv fd str start count [] in
	      if recvd = 0 then
		raise End_of_file
	      else
		recv_str fd str (start + recvd) (count - recvd)
    in
    let str_header = Bytes.create 10 in
      recv_str (self#in_fd ()) str_header 0 10;
      dprintf 10 ("network_buffer:    ..header (" ^
                    (Bytes.to_string str_header) ^ ") received..");
      let str_msg_len = int_of_string (Bytes.to_string str_header) in
      let str_msg = Bytes.create str_msg_len in
	recv_str (self#in_fd ()) str_msg 0 str_msg_len;
	dprintf 10 (Printf.sprintf "network_buffer:    ..body (%s) received:"
		      (Digest.to_hex (Digest.string (Bytes.to_string str_msg))));
	Marshal.from_string (Bytes.to_string str_msg) 0
end
