(* $Id: mucho_question.ml,v 1.20 2006/04/23 20:42:56 heinz Exp $
 *
 * GTF (Generic Test Framework): multiple choice question
 * Copyright (C) 2004  Heinz Deinhart
 *
 * This code is released under the GNU General Public License version 2 or
 * later.
 *)

open Gtf

type mucho_state =
    Yes | No | Unknown

class mucho_question csc (is_radio: bool) = object(self)
  inherit question csc

  val mutable question_text : string = ""
  val mutable answer_texts : string list = []
  val mutable solutions: bool array array = [||]
  val mutable state : mucho_state array = [||]
  val mutable is_radio = is_radio
  val mutable pictures: (string * string option) array = [||]

  val mutable number = 0

  method max_pts () =
    if is_radio then
      1
    else
      List.length answer_texts

  method min_pts () =
    (-1) * (self#max_pts ())

  method private dump_solutions () =
    dprintf 0 "solutions = [|";
    Array.iter (function ba ->
		  dprintf 0 "                [|";
		  Array.iter (function b -> dprintf 0 ("                    "^ (string_of_bool b))) ba;
		  dprintf 0 "                |]") solutions;
		  dprintf 0 "solutions = |]"

  method act_pts () =
    let cmpfun i b =
      match b, state.(i), is_radio with
	  true, Yes, _ | false, No, false -> 1
	| _, Unknown, _ | _, No, true -> 0
	| _, _, _ -> (-1)
    in 
      Array.fold_right max
	(Array.map (fun ba -> Array.fold_right (+) (Array.mapi cmpfun ba) 0)
	   solutions) min_int

  method classname () =
    "mucho_question"

  method archive () =
    Marshal.to_string (question_text, answer_texts, solutions, pictures,
		       is_radio, state) []

  method unarchive str =
      match Marshal.from_string str 0 with
	  (qt, at, s, pics, ir, _state) ->
	    question_text <- qt;
	    answer_texts <- at;
	    solutions <- s;
	    pictures <- pics;
	    is_radio <- ir;
	    state <- _state

  method status_normal () =
    let filter_fun s (u, nu) = if s = Unknown then (true, nu) else (u, true)
    in
      match Array.fold_right filter_fun state (false,false) with
	  true, false -> QS_Default
	| false, true -> QS_Done
	| _ -> QS_Modified

  method status_radio () =
    if Array.fold_right (fun s b -> (s == Yes) || b) state false then
      QS_Done
    else
      QS_Default

  method status () =
    if is_radio then self#status_radio () else self#status_normal ()

  method add_picture content name =
    pictures <- (Array.append pictures [| (content, name) |])

  method print (* explain *) _ =
    let (*ms2x*) _ = function Yes -> "Y" | No -> "N" | Unknown -> " "
    in
      "\n\n\\vspace{4mm}\n\\textbf{" ^ question_text ^ "}\\vspace{3mm}\n\n" ^
      "\\begin{tabular}{ccl}" ^
      (List.fold_right (fun s r -> r ^ s)
	 (List.rev
	    (List.map2
	       (fun str st ->
		  Printf.sprintf
		  "%s \\tablesep %s \\tablesep \\parbox[t]{12cm}{%s}\\\\\n"
		  (if st == Yes then "$\\bigotimes$" else "$\\bigcirc$")
  		  (if is_radio then "" else
		     (if st == No then "$\\bigotimes$" else "$\\bigcirc$"))
		  str)
	       answer_texts (Array.to_list state))) "") ^
       "\\rotatebox{90}{JA} \\tablesep \\rotatebox{90}{" ^
       (if is_radio then "" else "NEIN") ^ "} \\tablesep \\\n" ^
       "\\end{tabular}\n"

  method dump () =
    let ms2x = function Yes -> "Y" | No -> "N" | Unknown -> " "
    in
      dprintf 0 (Printf.sprintf "Q: %s\n" question_text);
      List.iter2 (fun str st ->
		    dprintf 0 (Printf.sprintf "  [%s] %s\n" (ms2x st) str))
	answer_texts (Array.to_list state)

  method register_number nr =
    number <- nr

  method populate_box packing =
    if is_radio then
      self#populate_box_radio packing
    else
      self#populate_box_normal packing

  method private populate_box_radio packing =
    let table_len = (List.length answer_texts) + 1
    in let hbox = GPack.vbox ~packing:packing ()
    in let table = GPack.table ~columns:3 ~rows:(table_len + 1)
		     ~col_spacings:2 ~row_spacings:6 ~packing:hbox#pack ()
    in let rec add_answers count row (buttons: GButton.toggle_button list) =
	function
	    str :: rest ->
	      let cbyes = GButton.check_button ~label:""
			    ~packing:(table#attach ~left:0 ~top:row;) ()
	      and _ = GMisc.label ~text:str ~line_wrap:true
			    ~packing:(GBin.alignment ~xalign:0.0 ~xscale:0.0
					~packing:(table#attach ~left:1
						    ~top:row) ())#add ()
	      in
		(GData.tooltips ())#set_tip cbyes#coerce ~text:"TRIFFT ZU";
    		add_answers (count + 1) (row + 1) (cbyes::buttons) rest
	  | [] ->
	      buttons
    in
    ignore (GMisc.label ~text:question_text ~line_wrap:true
	      ~packing:(GBin.alignment ~xalign:0.0 ~xscale:0.0
		          ~packing:(table#attach ~left:0 
                                      ~right:2 ~top:0) ())#add ());
    ignore (GMisc.label ~text:"JA" ~packing:(table#attach ~left:0 ~top:table_len) ());
    let buttons = Array.of_list (add_answers 0 1 [] answer_texts)
    in let pressed i () =
	 let b = buttons.(i)
	 in
	 if not b#active then
	   state.(i) <- No
	 else begin
	     Array.iteri (fun i2 b -> if i != i2 then b#set_active false)
	       buttons;
	     b#set_active true;
	     state <- Array.make (List.length answer_texts) No;
	     state.(i) <- Yes
	   end;
	 check_status_callback ()
       in
       Array.iteri (fun i (b: GButton.toggle_button) ->
	   if state.(i) = Yes then b#set_active true;
	   ignore (b#connect#toggled ~callback:(pressed i))) buttons

  method private populate_box_normal packing =
    let table_len = (List.length answer_texts)+1
    in let hbox = GPack.vbox ~packing:packing ()
    in let table = GPack.table ~columns:4 ~rows:(table_len + 1)
		     ~col_spacings:2 ~row_spacings:6 ~packing:hbox#pack ()
    in let rec add_answers count row =
	function
	    str :: rest ->
	      let cbyes = GButton.check_button ~label:""
			    ~packing:(table#attach ~left:0 ~top:row;) ()
	      and cbno = GButton.check_button ~label:""
			   ~packing:(table#attach ~left:1 ~top:row;) ()
	      and _ = GMisc.label ~text:str ~line_wrap:true
			    ~packing:(GBin.alignment ~xalign:0.0 ~xscale:0.0
					~packing:(table#attach ~left:2
						    ~top:row) ())#add ()
	      and cb_toggled this other newstate index () =
		begin
		  match this#active with
		      true ->
			other#set_active false;
			state.(index) <- newstate
		    | false ->
			state.(index) <- Unknown
		end;
		check_status_callback ()
	      in
		(GData.tooltips ())#set_tip cbyes#coerce ~text:"JA";
		(GData.tooltips ())#set_tip cbno#coerce ~text:"NEIN";
		ignore (cbyes#connect#toggled
                          ~callback:((cb_toggled cbyes cbno Yes) count));
		ignore (cbno#connect#toggled
                          ~callback:((cb_toggled cbno cbyes No) count));
		begin
		  match state.(count) with
		      Yes -> cbyes#set_active true
		    | No -> cbno#set_active true
		    | Unknown -> ()
		end;
		add_answers (count + 1) (row + 1) rest
	  | [] ->
	      ()
(*    in let add_image (content, name) packing =
	GMisc.pixmap ~packing:packing *)
    in
(*      Array.iter (add_image hbox#pack) pictures; *)
    ignore (GMisc.label ~text:question_text ~line_wrap:true
	      ~packing:(GBin.alignment ~xalign:0.0 ~xscale:0.0
		          ~packing:(table#attach ~left:0 ~right:3 ~top:0) 
                          ())#add ());
    ignore (GMisc.label ~text:"JA" ~packing:(table#attach ~left:0 ~top:table_len) ());
    ignore (GMisc.label ~text:"NEIN" ~packing:(table#attach ~left:1 ~top:table_len) ());
    add_answers 0 1 answer_texts

  method set_question_text str =
    question_text <- str;

  method add_answer text (solution: bool array) =
    let add_soluition (s: bool array) =
      if (Array.length solutions) = 0 then
	solutions <- Array.make (Array.length solution) [||]
      else
	if (Array.length solutions) != (Array.length s) then
	  raise (Invalid_argument "solution length");
      Array.iteri (fun i b -> solutions.(i) <-
		     Array.append solutions.(i) [| b |]) s
    in
      answer_texts <- List.append answer_texts [text];
      add_soluition solution;
      if is_radio then
	if (Array.length solution > 1) then
	  raise (Invalid_argument "solution length");
      state <- Array.append state [| if is_radio then No else Unknown |]

  method limit_answers count =
    let count = if is_radio || (count = 0) then (List.length answer_texts) else count
    in let iar = Array.make (List.length answer_texts) 0
       and at_ar = Array.of_list answer_texts
    in
      Array.iteri (fun i _ -> iar.(i) <- i) iar;
      let iar = Util.array_random_subarray iar count
      in
	answer_texts <- Array.to_list (Array.map (fun i -> at_ar.(i)) iar);
	state <- Array.map (fun i -> state.(i)) iar;
	solutions <- Array.map (fun ar ->
				  Array.map (fun i -> ar.(i)) iar) solutions
end
