(* $Id: server.ml,v 1.15 2018/07/30 12:52:20 heinz Exp $
 *
 * GTF (Generic Test Framework) Server
 * Copyright (C) 2004  Heinz Deinhart
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *)

open Gtf
(* open Batteries *)

let question_generator csc = function
    "mucho_question" -> (new Mucho_question.mucho_question csc false)#as_question ()
(*  | "word_question" -> (new Word_question.word_question csc)#as_question () *)
  | unknown -> raise (Invalid_argument ("Question class: " ^ unknown))

let die msg =
  Printf.printf "gtf: %s\ngtf: aborting\n" msg;
  exit 1

exception Usage of string option

let die_usage_death msg =
  begin
    match msg with
	None -> ignore ()
      | Some msg -> Printf.fprintf stderr "%s: %s\n" Sys.argv.(0) msg
  end;
  Printf.fprintf stderr "%s: usage: %s " Sys.argv.(0) Sys.argv.(0);
  output_string stderr "-p <port> [-b <bind-to-this-address-only>]\n";
  exit 1

let db_path = "./DB"

let dir_contents path =
  let dh = Unix.opendir path
  in let rec work result =
      try
	work ((Unix.readdir dh)::result)
      with
	  End_of_file -> result
  in
    work []

let db_get_dirs () =
  try
    let dh = Unix.opendir db_path
    in let rec work result =
	try
	  let f = Unix.readdir dh
	  in
	    if (Unix.stat (db_path^"/"^f)).Unix.st_kind = Unix.S_DIR then
	      if (f.[0] = '.') then
		work result
	      else
		work (f :: result)
	    else
	      work result
	with
	    End_of_file -> result
    in
      work []
  with
      Unix.Unix_error (Unix.ENOENT, "opendir", db_path) ->
	Printf.fprintf stderr "%s: DB %s not found.\n" Sys.argv.(0) db_path;
	[]
    | _ ->
	Printf.fprintf stderr "%s: db_get_dirs: unknown error\n" Sys.argv.(0);
	[]

let db_get_entries dir =
  try
    List.filter (fun f -> (Unix.stat f).Unix.st_kind = Unix.S_REG)
      (dir_contents (db_path ^ "/" ^ dir))
  with
      _ ->
	Printf.fprintf stderr "%s: db_get_dirs failed\n" Sys.argv.(0);
	[]

class virtual selector = object(self)
  method virtual register : Unix.file_descr -> (unit -> unit) -> unit
  method virtual unregister : Unix.file_descr -> unit
  method as_selector () =
    (self :> selector)
end

(* returns string option
 *)
let lookup_test_info dir =
  try
    let filename = db_path ^ "/" ^ dir ^ "/info.txt"
    in let ich = open_in_bin filename
    in let buf = Bytes.create (in_channel_length ich)
    in
      really_input ich buf 0 (in_channel_length ich);
      Some (Bytes.to_string buf)
  with
      _ -> None

let lookup_test dir idnr lastname =
  let basename = db_path ^ "/" ^ dir ^ "/" ^ idnr
  in let fname = if Sys.file_exists (basename ^ ".gtfr") then
      raise (Invalid_argument "test already completed")
    else
      if Sys.file_exists (basename ^ ".gtfi") then
	(basename ^ ".gtfi")
      else
	(basename ^ ".gtft")
  in let ich =  open_in fname
     and s = new Gtf.student
  in let (student_archive, test_archive) = Marshal.from_channel ich
  in let _ = new Gtf.test question_generator
  in
    dprintf 1 ("loaded test file " ^ fname);
    s#unarchive student_archive;
    if s#compare_idnr idnr then
      if s#compare_lastname lastname then
	test_archive
      else
	raise (Invalid_argument "no such lastname")
    else
      raise (Invalid_argument "no such idnr")

class session sock the_selector = object(self)
  inherit network_buffer

  val socket = sock
  val mutable selector = the_selector

  initializer
    selector#register socket self#handle_msg

  method in_fd () =
    socket

  method out_fd () =
    socket

  method handle_msg () =
    try
      match self#recv_msg () with
	  RequestDirectory ->
	    self#send_msg (Directory (db_get_dirs ()))
	| RequestEntries dir ->
	    self#send_msg (Entries (dir, (db_get_entries dir)))
	| RequestTest (dir, idnr, lastname) ->
	    begin
	      try
		self#send_msg (Test ((lookup_test_info dir), dir, idnr,
				     (lookup_test dir idnr lastname)))
	      with
		  Sys_error msg
		| Invalid_argument msg ->
		    dprintf 0 ("failed to open test: " ^ msg);
		    self#send_msg (ServerError ("failed to open test: " ^ msg))
	    end
	| Finish (dir, idnr, (stud_archive, test_archive)) ->
	    begin
	      try
		let fname = db_path ^ "/" ^ dir ^ "/" ^ idnr ^ ".gtfr"
		in let och = open_out fname
		in
		  Marshal.to_channel och (stud_archive, test_archive) [];
		  close_out och
	      with
		  x ->
		    dprintf 0 (Printf.sprintf
				 "failed to save intermediate test %s: %s"
				 db_path (Printexc.to_string x));
		    self#send_msg (ServerError
				     ("failed to save finished test"))
	    end;
	    self#send_msg (Directory (db_get_dirs ()))
	| Intermediate (dir, idnr, (stud_archive, test_archive)) ->
	    begin
	      try
		let fname = db_path ^ "/" ^ dir ^ "/" ^ idnr ^ ".gtfi"
		in let och = open_out fname
		in
		  Marshal.to_channel och (stud_archive, test_archive) [];
		  close_out och
	      with
		  x ->
		    dprintf 0 (Printf.sprintf
				 "failed to save intermediate test %s: %s"
				 db_path (Printexc.to_string x));
		    self#send_msg (ServerError
				     ("failed to save intermediate test"))
	    end
	| ClientError str ->
	    Printf.fprintf stderr "%s: received ClientError '%s'.\n"
	    Sys.argv.(0) str;
	| _ ->
	    raise GTF_protocol_error
    with
	_ ->
	  dprintf 0 "msg read failed, killing session";
	  selector#unregister socket;
	  Unix.close socket
end

class listener bind_addr port (selector: selector) = object(self)
  val socket = Unix.socket Unix.PF_INET Unix.SOCK_STREAM 0
  val the_selector = selector

  initializer
  let sock_addr = Unix.ADDR_INET(bind_addr, port)
  in
    Unix.setsockopt socket Unix.SO_REUSEADDR true;
    Unix.bind socket sock_addr;
    Unix.listen socket 5;
    selector#register socket self#accept

  method accept () =
    try
      let (new_socket, _) = Unix.accept socket
      in
	dprintf 5 "i just accepted a new connection";
	ignore (new session new_socket (the_selector#as_selector ()))
    with x -> dprintf 0 (Printf.sprintf "listener::accept failed: %s"
			   (Printexc.to_string x))
end

let hashtbl_keys h =
  Hashtbl.fold (fun k _ c -> k::c) h []

class gtf_server = object(self)
  inherit selector

  val sockets = Hashtbl.create 10

  method run bind_addr (port: int) =
    ignore (new listener bind_addr port (self#as_selector ()));
    self#main_loop ()

  method register fd callback =
    dprintf 10 "registering select callback";
    Hashtbl.add sockets fd callback

  method unregister fd =
    Hashtbl.remove sockets fd

  method main_loop () =
    let rec forever () =
      let r, _, _ = Unix.select (hashtbl_keys sockets) [] [] (-1.0)
      in
	List.iter (function fd -> (Hashtbl.find sockets fd)()) r;
	forever ()
    in
      ignore (forever ())
end

let rec parse_args argqueue port bind_addr =
  if not (Queue.is_empty argqueue) then
    try
      match Queue.pop argqueue with
	  "-p" ->
	    parse_args argqueue (int_of_string (Queue.pop argqueue)) bind_addr
	| "-b" ->
	    parse_args argqueue port (Unix.inet_addr_of_string
					(Queue.pop argqueue))
	| "-v" ->
	    dprintf 0 (Printf.sprintf "%s: version 0.2.0" Sys.argv.(0));
	    exit 0
	| _ ->
	    raise Queue.Empty
    with
	Queue.Empty ->
	  raise (Usage None)
      | Failure str ->
	  raise (Usage (Some ("illegal argument: " ^ str)))
      | _ ->
	  raise (Usage (Some "unknown error"))
  else
    (port, bind_addr)

let _ =
  try
    let argqueue = Queue.create ()
    in
      Array.iter (fun s -> Queue.push s argqueue) Sys.argv;
      ignore (Queue.pop argqueue); (* discard argv[0] *)
      let port, bind_addr = parse_args argqueue 0 Unix.inet_addr_any
      in
	if port = 0 then
	  raise (Usage (Some "argument -p is mandatory"));
	(new gtf_server)#run bind_addr port
  with
      Usage msg ->
	die_usage_death msg
